/************************************************************************/
/* sos.c                                                                */
/*                                                                      */
/* A successive orders of scattering  model for radiative transfer.     */
/*                                                                      */
/* Author: Arve Kylling,                                                */
/*         NILU, arve.kylling@nilu.no                                   */
/************************************************************************/


#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <time.h>
#include "ascii.h"
#include "../src_f/f77-uscore.h"
#include "fortran_and_c.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

void F77_FCN (qgausn) (int* n, float* cmu, float*cwt);
void F77_FCN (lepoly) (int* nn, int* mazim, int* mxcmu, 
		       int* nstr, float* cmu, float* ylmc);

float chapman(int lc, int *nfac, float **fac, float *dtauc ) {

  /*     Calculates the Chapman-factor, Eq. B2 (DS).                     */

  /* I n p u t       v a r i a b l e s:                                  */
   
  /*      lc        : Computational layer                                */
  /*      dtauc     : Optical thickness of layer lc (un-delta-m-scaled)  */
  /*      fac       : Geomtrical correction factor, calculated by routine*/
  /*                  geofac.                                            */
  /*      nfac      : Number of terms in Eq. B2 (DS)                     */
   
  /* O u t p u t      v a r i a b l e s:                                 */
   
  /*      chapman   : Chapman-factor. In a pseudo-spherical atmosphere   */
  /*                  replace EXP( -tau/umu0 ) by EXP( -chapman ) in the */
  /*                  beam source.                                       */
   
  /* I n t e r n a l     v a r i a b l e s:                              */
   
  /*      fact      : =1 for first sum in Eq. B2 (DS)                    */
  /*                  =2 for second sum in Eq. B2 (DS)                   */
   
  float fact, sum=0;
  int j;  

  sum = 0.0;
  if ( nfac[lc] < 0 ) { sum = 1.e+20; return sum; }
  else {
    for (j=0;j<=nfac[lc];j++) {
      /* Include factor of 2 for zenang .gt. 90, second sum in Eq. B2 (DS) */
      fact = 1.0;
      if( j > lc ) {fact = 2.0;}
      sum += dtauc[j]*fact*fac[lc][j];
    }
    if ( nfac[lc] > lc ) {
      sum += dtauc[lc] * fac[lc][j+1];
    }
  }
  return sum;
}

int geofac(float **fac, int *nfac, float z_lay, int nlyr,
	   float *zd, float zenang, float r )  
{
  /* Calculates the geometric correction factor needed for spherical      */
  /* geometry. See Appendix B, (DS).                                      */

  /* I n p u t       v a r i a b l e s:                                   */

  /*      nlyr      : Number of layers in atmospheric model               */
  /*      zd(lc)    : lc = 0, nlyr. zd(lc) is distance from bottom        */
  /*                  surface to top of layer lc. zd(nlyr) = 0.0 km       */
  /*      zenang    : Solar zenith angle as seen from bottom surface      */ 
  /*      r         : Radius of earth. NOTE: Use the same dimension as zd,*/
  /*                  for instance both in km.                            */
  /*      z_lay     : Where in the layer the Chapman function is to be    */
  /*                  computed. E.g. 0.0, bottom of layer, 0.5, middle    */ 
  /*                  of layer and 1.0 top of layer.                      */

  /* O u t p u t      v a r i a b l e s:                                  */
     
  /*      fac        : geometrical correction factor                      */
  /*      nfac       : Number of terms in Eq. B2 (DS)                     */

  /* I n t e r n a l     v a r i a b l e s:                               */

  /*      dhj       : delta-h-sub-j in Eq. B2 (DS)                        */
  /*      dsj       : delta-s-sub-j in Eq. B2 (DS)                        */
  /*      fact      : =1 for first sum in Eq. B2 (DS)                     */
  /*                  =2 for second sum in Eq. B2 (DS)                    */
  /*      rj        : r-sub-j in Eq. B1 (DS)                              */
  /*      rjp1      : r-sub-j+1 in Eq. B1 (DS)                            */
  /*      xpsinz    : The length of the line OG in Fig. 1, (DS)           */

  float dsj, dhj, fact, rj, rjp1, xp, xpsinz, zenrad;
  int id, j, lc, status =0;

  zenrad = zenang * PI / 180.0;
  for (lc=0;lc<nlyr;lc++) {
    nfac[lc] = 0;
  }

  for (lc=0;lc<nlyr;lc++) {
    xp     = r +  zd[lc+1] + (zd[lc] - zd[lc+1] ) * z_lay;
    xpsinz = xp * sin( zenrad );
    if( (zenang > 90.0) && (xpsinz < r) ) {
      nfac[lc] = -1;
    }
    else {
      /*  Find index of layer in which the screening height lies */
      id = lc;
      if ( zenang > 90.0 ) {
	for (j=lc;j<nlyr;j++) {
	  if( (xpsinz < ( zd[j] + r ) ) &&
	      (xpsinz >= ( zd[j+1] + r )) ) {id = j;}
	}
      }
      for (j=0;j<=id;j++) {
	fact = 1.0;
	/* Include factor of 2 for zenang .gt. 90, second sum in Eq. B2 (DS)*/
	if( j > lc ) { fact = 2.0;}
	if(j==id && id==lc && zenang>90.0) { fact = -1.0;}
	rj = r + zd[j];
	rjp1 = r + zd[j+1];
	if(j==lc && id==lc) {rjp1 = xp;}
	dhj = zd[j] -zd[j+1];
	if (id>lc && j==id) { dsj = sqrt(rj*rj - xpsinz*xpsinz );}
	else { dsj = sqrt( rj*rj - xpsinz*xpsinz ) -
		 fact * sqrt( rjp1*rjp1 - xpsinz*xpsinz );}
	fac[lc][j] = dsj / dhj;
      }
      nfac[lc] = id;
     
      /* Third term in Eq. B2 (DS) */
      if( id > lc ) {
	dhj = zd[lc] -zd[lc+1];
	dsj = sqrt( xp*xp - xpsinz*xpsinz ) -
	  sqrt( (zd[lc+1]+r)*(zd[lc+1]+r) - xpsinz*xpsinz );
	fac[lc][j+1] = dsj / dhj;
      }
    }
  }
  return status;
}

float sinsca(float* ssalb, float* tauc, float utau, float* chtau, int layru,
	     int iq,
	     float **pmom, float cmu, float* cwt, int nstr, float **ylmc, float **ylm0) {
  int lc, l;
  float dtau=0,sum=0, phase=0,part;
  if (cmu < 0) {
    sum =0.;
  }
  else if (cmu > 0) {
    sum=0;
    for (lc=0;lc<=layru;lc++) {
      phase = 0;
      for (l=0;l<nstr;l++) {
	part = (2*l+1) * ylmc[iq][l] * ylm0[0][l] * pmom[lc][l] * cwt[l];
	phase += (2*l+1) * ylmc[iq][l] * ylm0[0][l] * pmom[lc][l] * cwt[l] * 0.5;
		printf("sin %d %d %f %f %f %f %f %f\n",l, lc, phase, ylmc[iq][l] , ylm0[0][l] , pmom[lc][l], part, cwt[l]);
      }
      dtau = tauc[lc+1]-tauc[lc];
      dtau = dtau/cmu;
      sum += ssalb[lc]*exp(-chtau[lc])*exp(-(utau-tauc[lc])/cmu)*dtau;
    }
    sum *= phase/4;
    printf("sum:  %d %f %f\n", lc, phase, sum);
  }
  return sum;
}

float morsca(float* ssalb, float* tauc, float utau, float* chtau, int layru,
	     int is, int iq, float **pmom, float cmu, float *cwt, int nstr,
	     float **ylmc, float **ylm0, float ***u0uc ) {
  int jq, lc, l;
  float dtau=0,sum=0, phase=0, source=0;
  if (cmu < 0) {
    sum =0.;
  }
  else if (cmu > 0) {
    sum=0;
    for (lc=0;lc<=layru;lc++) {
	printf("ssalbd morsca:  %d %f\n", lc,  ssalb[lc]);
      source = 0;
      for (jq=0;jq<nstr;jq++) {
	phase = 0;
	for (l=0;l<nstr;l++) {
	  phase += (2*l+1) * ylmc[iq][l] * ylmc[jq][l] * pmom[lc][l] * cwt[l];
	  /*	  printf("%d %f %f %f %f\n",
		  l,ylmc[iq][l], ylmc[jq][l], pmom[lc][l], phase);*/
	}
	source += phase*u0uc[is][lc][jq]*cwt[jq];
	printf("source %d %d %d %f %f\n",is,lc,jq,u0uc[is][lc][jq], source);
      }
      dtau = tauc[lc+1]-tauc[lc];
      dtau = dtau/cmu;
      sum += source*exp(-chtau[lc])*exp(-(utau-tauc[lc])/cmu)*dtau*ssalb[lc];
      printf("source: %d %e %e %f\n",lc,phase,source,ssalb[lc]);
    }
    sum *= phase*0.5;
    printf("sum:  %d %f %f %f\n", lc, phase, sum, ssalb[lc]);
  }
  return sum;
}

int sos (int nlyr, int newgeo, int nstr, int nscat,
	 float radius, float *zd, float *ssalb, float **pmom,
	 float *dtauc, float zenang, int ntau, float *utau,
	 float *rfldir, float *rfldn,  float *flup,
	 float *uavgso, float *uavgdn, float *uavgup)
{
  float cumtau=0;
  float umu0, /* *umu ,*/ **fac=NULL, *chtau=NULL, *tauc=NULL;
  float *cmu=NULL, *cwt=NULL, /* ***uuc ,*/ *tmpylmc, 
    **ylmc=NULL, **ylm0=NULL, ***u0uc, sgn;
  float *tmp;
  int *layru=NULL, *nfac=NULL;
  int status=0, mazim, twonm1;
  int iq, is, l, lc, lu, nn;

  /* Do some initialization stuff */
  umu0 = cos(zenang/180.0*PI);
  if ( newgeo ) {
    /* Calculate geometric correction factor needed for chapman function */
    /* Only need to do this if the geometry changes, that is, it is not  */
    /* wavelength dependent. */
    if ( (layru = (int *) calloc ((size_t) (ntau), sizeof (int))) == NULL )  
      return ASCII_NO_MEMORY;
    if ( (nfac = (int *) calloc ((size_t) (nlyr), sizeof (int))) == NULL )  
      return ASCII_NO_MEMORY;
    if ( (tauc = (float *) calloc ((size_t) (nlyr+1), sizeof (float))) == NULL )
      return ASCII_NO_MEMORY;
    if ((status = ASCII_calloc_float (&fac, nlyr, 2*nlyr)) != 0) return status;
    status = geofac( fac, nfac, 0.0, nlyr, zd, zenang, radius ) ;
    if (status!=0) {
      fprintf (stderr, "error %d during geofac\n", status);
      return status; 
    }
    if ((status = ASCII_calloc_float_3D (&u0uc, nscat, nlyr, nstr)) != 0)
      return status;
    chtau = calloc ((size_t) (nlyr+1), sizeof(float));
    if ( (cmu = (float *) calloc ((size_t) (nstr), sizeof (float))) == NULL )
      return ASCII_NO_MEMORY;
    if ( (cwt = (float *) calloc ((size_t) (nstr), sizeof (float))) == NULL )
      return ASCII_NO_MEMORY;
    nn = nstr/2;
    F77_FCN (qgausn) (&nn,cmu,cwt);
    /* Rearrange to get cmu in ascending order */
    for (iq=0;iq<nn;iq++) { cmu[nn+iq] = cmu[iq]; cwt[nn+iq] = cwt[iq]; }
    for (iq=0;iq<nn;iq++) { cmu[iq] = -cmu[nstr-1-iq]; cwt[iq] = cwt[nstr-1-iq];}
    
    if ( (tmpylmc = (float *) calloc ((size_t) ((nstr+1)*nstr), sizeof (float))) == NULL )
      return ASCII_NO_MEMORY;
    mazim=0;
    twonm1=nstr-1;
    nn=1;
    if ( (tmp = (float *) calloc ((size_t) (1), sizeof (float))) == NULL )
      return ASCII_NO_MEMORY;
    tmp[0] = -umu0;
    F77_FCN (lepoly) ( &nn, &mazim, &nstr, &twonm1, tmp, tmpylmc );
    free(tmp);
    ylm0 = fortran2c_2D_float_ary(nstr, nstr+1, tmpylmc);
    nn=nstr/2;
    F77_FCN (lepoly) ( &nn, &mazim, &nstr, &twonm1, cmu, tmpylmc );
    ylmc = fortran2c_2D_float_ary(nstr, nstr+1, tmpylmc);
    free(tmpylmc);
    /* Evaluate Legendre polynomials with negative -cmu- from those with*/
    /* positive -cmu-;  Dave Armstrong Eq. (15) */
     for (l=0;l<nstr;l++) { 
       for (iq=0;iq<nstr;iq++) { 
 	printf("a lepoly %d %d %f %f\n",iq,l,ylmc[iq][l],cmu[iq]); 
       } 
     }     
    sgn  = -1.0;
    for(l=0;l<nstr;l++) {
      sgn = -sgn;
      for (iq=nn;iq<nstr;iq++) {
	 ylmc[iq][l] = sgn*ylmc[nstr-1-iq][l];
 	printf("c lepoly %d %d %d %f %f\n",l,iq,nstr-1-iq,ylmc[iq][l],ylmc[nstr-1-iq][l]); 
      }
    }
/*     for (iq=0;iq<nstr;iq++) { */
/*       printf("cmu %d %f %f\n",iq,cmu[iq],cwt[iq]); */
/*     } */
     for (l=0;l<nstr;l++) { 
       for (iq=0;iq<nstr;iq++) { 
 	printf("b lepoly %d %d %f %f\n",l,iq,ylmc[iq][l],cmu[iq]); 
       } 
     }     
/*     for (l=0;l<nstr;l++) {  */
/*       for (iq=0;iq<nstr;iq++) {  */
/*  	printf("lepoly %d %d %f %f\n",iq,l,ylm0[iq][l],cmu[iq]);  */
/*       }  */
/*     }      */

  }

  /* Chapman factor */
  cumtau = 0;
  tauc[0]=0;
  chtau[0]=0;
  for (lc=0;lc<nlyr;lc++) {
    chtau[lc+1] = chapman(lc, nfac, fac, dtauc);
    cumtau += dtauc[lc];
    tauc[lc+1] = cumtau;
  }

  /* Set arrays defining location of user output levels */
  for (lu=0;lu<ntau;lu++) {
    for (lc=0;lc<nlyr;lc++) {
      if ( utau[lu]>=tauc[lc] && utau[lu]<=tauc[lc+1] )
	goto done; /* All good programs need at least one goto statement:))*/
    }
    lc = nlyr-1;
  done:
    layru[lu] = lc;
  }

  /* The real stuff for user defined order of scattering */

  /* Single scattering */
  if (nscat >=1) {
    is = 0;
    printf("nlyr: %d\n",nlyr);
    for (lc=0;lc<nlyr;lc++) {
      for (iq=0;iq<nstr;iq++) {
	u0uc[is][lc][iq] = sinsca(ssalb, tauc, tauc[lc], chtau, lc, iq,
				  pmom, cmu[iq], cwt, nstr, ylmc, ylm0);
	printf("sin u0uc %d %d %d %e\n",is,lc,iq,u0uc[is][lc][iq]);
      }
    }
  }

  /* Higher order scattering */
  if (nscat > 1) {
    for (is=1;is<nscat;is++) {
      for (lc=0;lc<nlyr;lc++) {
	printf("ssalbd:  %d %f\n", lc,  ssalb[lc]);
	for (iq=0;iq<nstr;iq++) {
	  u0uc[is][lc][iq] = morsca(ssalb, tauc, tauc[lc], chtau, lc, is-1, iq,
				    pmom, cmu[iq], cwt, nstr, ylmc, ylm0, u0uc);
	  printf("mor u0uc %d %d %d %f\n",is,lc,iq,u0uc[is][lc][iq]);
	}
      }
    }
  }

  /* Output at user levels */
  for (lu=0;lu<ntau;lu++) {
    rfldn[lu] = 0;
    for (is=0;is<nscat;is++) {
      for (iq=nn;iq<nstr;iq++) {
	rfldn[lu] += cwt[iq]*u0uc[is][layru[lu]][iq];
	printf("rfldn %d %d %d %d %f %f %f\n",is,lc,iq,layru[lu],rfldn[lu],cwt[iq],u0uc[is][layru[lu]][iq]);
      }
      printf("rfldn %d %f\n",is,rfldn[lu]);
    }
    rfldir[lu] = umu0 * exp(-chtau[layru[lu]+1]);
    printf("rfldir %f\n",rfldir[lu]);
  }
  return status;
}
