/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include <stdio.h>							    
#include <stdlib.h>
#include <math.h>
#include <ckdfu.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_f/f77-uscore.h"

#define EPSILON 1E-6


/************************************/
/* prototypes of internal functions */
/************************************/

static int shift_cloud (atm_inp_struct *atm_inp, 
			cld_out_struct  *cld_out);

static int compare_levels (float *z1, int n1,
			   float *z2, int n2);




/**************************************************************/
/* Setup water clouds.                                        */
/**************************************************************/

int setup_wcloud (input_struct input, output_struct *output)
{
  int status=0, lev=0, iv=0, rows=0, min_columns=0, max_columns=0;

  float **wc_data=NULL;
  float sum=0, scale_factor=0;

  int nstring = (int) strlen(input.filename[FN_PATH]);

  void  F77_FCN (wcloud) (float* lambda_t, int* newsiz, int* nlyr,
			  char* filepath, int* nstring, float* lwc, float* wceffr, 
			  float* tmp_wc_dtau, float* tmp_wc_gg, float* tmp_wc_ssa, float* zd,
			  int *wclyr);


  if ( strlen(input.filename[FN_WCLOUD]) > 0 ) {

    status = ASCII_file2float (input.filename[FN_WCLOUD], 
			       &rows, &max_columns, &min_columns, 
			       &wc_data);
    if (status!=0) {
      fprintf (stderr, " ... error %d opening file %s\n", status, input.filename[FN_WCLOUD]);
      return status;
    }

    if (max_columns!=min_columns) {
      fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    }  

    if (min_columns<3) {
      fprintf (stderr, " ... error, found less than 3 columns in %s\n",
	       input.filename[FN_WCLOUD]);
      return -1;
    }


    /* allocate memory for 1D cloud structure, output->wc */
    status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, rows); 
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }

    for (lev=0; lev<output->wc.nlev; lev++) {
      output->wc.zd  [lev] = wc_data[lev][WCZD];
      output->wc.lwc [lev] = wc_data[lev][WCLWC];
      output->wc.effr[lev] = wc_data[lev][WCEFFR];

      if (output->wc.lwc [lev] > 0) {
	if (output->wc.effr[lev] < 2.5 || output->wc.effr[lev] > 60.0) {
	  fprintf (stderr, "Error: effective droplet radius out of range:\n");
	  fprintf (stderr, "       level = %d, reff = %g, outside allowed range %g - %g\n",
		   lev, output->wc.effr[lev], 2.5, 60.0);
	  return -1;
	}
      }
    }

    /* Shift clouds to given altitude */
    if (input.atm.altitude != 0) {
      status = shift_cloud(&input.atm, &output->wc);
      if (status !=0) {
	fprintf (stderr, "error %d shifting clouds\n", status);
	return status;
      }
    }

    /* check for consistency with atmospheric levels */
    status = compare_levels (output->wc.zd, output->wc.nlev,
			     output->atm.zd, output->atm.nlev);
    
    if (status!=0) {
      fprintf (stderr, "Error: z-grids of atmosphere and 1D clouds are different\n");
      return status;
    }
  
    /* newsiz is a flag, telling the Fortran function wcloud() if a */
    /* new liquid water content and/or effective radius are given   */
    /* (newsiz=1) or if the values haven't changed since the last   */
    /* call (newsiz=0).                                             */ 
    output->wc.newsiz = 1;


    /* calculate cloud optical properties */
    switch (input.ck_scheme) {
    case CK_FU:

      /* call Fu and Liou code by Fred Rose */

#if HAVE_FULIOU
      status = ckdfucld (output->wc.effr, output->wc.lwc, output->atm.nlev,
			 output->wc.dtau, output->wc.gg, output->wc.ssa);
#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif

      if (status!=0) {
	fprintf (stderr, "Error %d returned by ckdfucld()\n", status);
	return status;
      }

      break;

    default:

      for (iv=0; iv<output->wl.nlambda_t; iv++) {
	F77_FCN (wcloud) (&output->wl.lambda_t[iv], &output->wc.newsiz, &output->atm.nlyr,
			  input.filename[FN_PATH], &nstring, output->wc.lwc, output->wc.effr, 
			  output->wc.dtau[iv], output->wc.gg[iv], output->wc.ssa[iv], 
			  output->atm.zd, &input.wc.layer);
      }

      break;
    }


    for (iv=0; iv<output->wl.nlambda_t; iv++) {

      /**** Set the water cloud optical depth to a constant user specified value ****/
      if (input.wc.tau >= 0.0) {
	sum = 0.0;
	for (lev=0;lev<output->atm.nlyr;lev++)
	  sum += output->wc.dtau[iv][lev];
	
	if (sum>0)
	  scale_factor = input.wc.tau/sum;
	else 
	  scale_factor = 0.0;

	for (lev=0; lev<output->atm.nlyr; lev++)
	  output->wc.dtau[iv][lev] *= scale_factor;
      }


      /**** Set the water cloud asymmetry factor to a user specified value ****/
      if ( input.wc.gg >= -1.0 )
	for (lev=0;lev<output->atm.nlyr;lev++)
	  output->wc.gg[iv][lev] = input.wc.gg;
      
      /**** Set the water cloud single scattering albedo to a user specified value ****/
      if ( input.wc.ssa >= 0.0 )
	for (lev=0;lev<output->atm.nlyr;lev++)
	  output->wc.ssa[iv][lev] = input.wc.ssa;
      
    }

    /* free memory */
    status = ASCII_free_float(wc_data, rows);
  }
  else {   
    /* allocate memory, even if no cloud is defined; */
    /* it may be required by other functions.        */

    status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, output->atm.nlev);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
      return status;
    }
  }
  
  /*
  iv=0;
  for (lev=0;lev<output->atm.nlyr;lev++) {
    if (output->wc.dtau[iv][lev]>0)
      fprintf (stdout, "wat %d %d  %g %g  %g %g %g %g\n", 
	       iv, lev, output->wc.zd[lev], 
	       output->wc.lwc[lev], output->wc.effr[lev],
	       output->wc.dtau[iv][lev], output->wc.ssa[iv][lev], output->wc.gg[iv][lev]);
  }
  */

  return 0;
}




/**************************************************************/
/* Setup ice clouds.                                          */
/**************************************************************/

int setup_ice (input_struct input, output_struct *output)
{
  int status=0, lev=0, iv=0, rows=0, min_columns=0, max_columns=0;

  float **ic_data=NULL;
  float sum=0, scale_factor=0;


  if ( strlen(input.filename[FN_ICLOUD]) > 0 ) {

    status = ASCII_file2float (input.filename[FN_ICLOUD], 
			       &rows, &max_columns, &min_columns, 
			       &ic_data);
    if (status!=0) {
      fprintf (stderr, " ... error %d opening file %s\n", status, input.filename[FN_ICLOUD]);
      return status;
    }

    if (max_columns!=min_columns) {
      fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    }  

    if (min_columns<3) {
      fprintf (stderr, " ... error, found less than 3 columns in %s\n",
	       input.filename[FN_ICLOUD]);
      return -1;
    }


    /* allocate memory for 1D cloud structure, output->ic */
    status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, rows); 
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for ic_out_struct\n", status);
      return status;
    }

    for (lev=0; lev<output->ic.nlev; lev++) {
      output->ic.zd  [lev] = ic_data[lev][ICZD];
      output->ic.lwc [lev] = ic_data[lev][ICLWC];
      output->ic.effr[lev] = ic_data[lev][ICEFFR];
    }

    /* Shift clouds to given altitude */
    if (input.atm.altitude != 0) {
      status = shift_cloud(&input.atm, &output->ic);
      if (status !=0) {
	fprintf (stderr, "error %d shifting clouds\n", status);
	return status;
      }
    }

    /* check for consistency with atmospheric levels */
    status = compare_levels (output->ic.zd, output->ic.nlev,
			     output->atm.zd, output->atm.nlev);
    
    if (status!=0) {
      fprintf (stderr, "Error: z-grids of atmosphere and 1D clouds are different\n");
      return status;
    }
  
    
    /* calculate cloud optical properties */
    switch (input.ck_scheme) {
    case CK_FU:
      
#if HAVE_FULIOU
      status = ckdfuice (output->ic.effr, output->ic.lwc, output->atm.nlev,
			 output->ic.dtau, output->ic.gg, output->ic.ssa);
#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif

      if (status!=0) {
	fprintf (stderr, "Error %d returned by ckdfuice()\n", status);
	return status;
      }

      break;

    default:

      fprintf (stderr, "Error, ice clouds only implemented for FU parameterization\n");
      return -1;
      break;
    }



    for (iv=0; iv<output->wl.nlambda_t; iv++) {

      /**** Set the ice cloud optical depth to a constant user specified value ****/
      if (input.ic.tau >= 0.0) {
	sum = 0.0;
	for (lev=0;lev<output->atm.nlyr;lev++)
	  sum += output->ic.dtau[iv][lev];
	
	if (sum>0)
	  scale_factor = input.ic.tau/sum;
	else 
	  scale_factor = 0.0;

	for (lev=0; lev<output->atm.nlyr; lev++)
	  output->ic.dtau[iv][lev] *= scale_factor;
      }

      /**** Set the ice cloud asymmetry factor to a user specified value ****/
      if ( input.ic.gg >= -1.0 )
	for (lev=0;lev<output->atm.nlyr;lev++)
	  output->ic.gg[iv][lev] = input.ic.gg;
      
      /**** Set the ice cloud single scattering albedo to a user specified value ****/
      if ( input.ic.ssa >= 0.0 )
	for (lev=0;lev<output->atm.nlyr;lev++)
	  output->ic.ssa[iv][lev] = input.ic.ssa;
      
    }

    /* free memory */
    status = ASCII_free_float(ic_data, rows);
  }
  else {   
    /* allocate memory, even if no cloud is defined; */
    /* it may be required by other functions.        */

    status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, output->atm.nlev);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for ic_out-struct\n", status);
      return status;
    }
  }
  

  /*
  iv=0;
  for (lev=0;lev<output->atm.nlyr;lev++) {
    if (output->ic.dtau[iv][lev]>0)
      fprintf (stdout, "ice %d %d  %g %g  %g %g %g %g\n", 
	       iv, lev, output->ic.zd[lev], 
	       output->ic.lwc[lev], output->ic.effr[lev],
	       output->ic.dtau[iv][lev], output->ic.ssa[iv][lev], output->ic.gg[iv][lev]);
  }
  */

  return 0;
}




/**************************************************************/
/* Shift clouds the same way as the atmosphere                */
/* (see shift_atmosphere).                                    */
/**************************************************************/

static int shift_cloud (atm_inp_struct *atm_inp, 
			cld_out_struct  *cld_out)
{
  int lev=0;

  /* calculate shifted levels (altitude now in km) */
  for (lev=0; lev<cld_out->nlev; lev++)
    cld_out->zd[lev] = cld_out->zd[lev] - atm_inp->altitude;

  /* check for strange things */
  for (lev=0; lev<cld_out->nlev; lev++)
    if (cld_out->zd[lev]<0 && cld_out->lwc[lev]>0) {
      fprintf (stderr, "Error: the first cloud level is below the surface; you need\n");
      fprintf (stderr, "to change the cloud profile!\n");
      return -1;
    }


  /* determine the first level below 0 */
  lev = cld_out->nlev;
  while (cld_out->zd[--lev]<=0) {
    if (lev<=1) {
      fprintf (stderr, "Sorry, ALTITUDE is higher than the uppermost layer in the\n");
      fprintf (stderr, "water cloud profile.\n");
      return -1;
    }
  }

  /* number of levels is probably smaller now */
  cld_out->nlev = lev+2;   /* ??? */

  /* set the last level to 0 */
  cld_out->zd[cld_out->nlev-1] = 0.0; 


  return 0;
}




/*********************************/
/* compare two altitude grids    */
/*********************************/

static int compare_levels (float *z1, int n1,
			   float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "error, number of layers differing\n");
    fprintf (stderr, "%d in 3D, %d in 1D\n", n1, n2);

    return -1;
  }

  for (kc=0; kc<n1; kc++)
    if (fabs((float) (z1[kc]) - z2[kc]) > EPSILON * z1[kc]) {
      fprintf (stderr, "error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;  /* if o.k. */
}





/***************************************/
/* allocate memory for a cld_out_struct */
/***************************************/

int calloc_cld_out (cld_out_struct *out, int Nlam, int Nlev) 
{
  int iv=0;

  out->dtau = calloc(Nlam, sizeof (float *));
  out->gg   = calloc(Nlam, sizeof (float *));
  out->ssa  = calloc(Nlam, sizeof (float *));
    
  out->nlev = Nlev;
  out->nlam = Nlam;
  out->newsiz = 1;
  
  out->zd   = calloc(Nlev, sizeof(float));
  out->lwc  = calloc(Nlev, sizeof(float));
  out->effr = calloc(Nlev, sizeof(float));

  for (iv=0; iv<Nlam; iv++) {
    out->dtau[iv] = calloc (Nlev, sizeof(float));
    out->gg  [iv] = calloc (Nlev, sizeof(float));
    out->ssa [iv] = calloc (Nlev, sizeof(float));
  }

  return 0;
}





/************************************************/
/* copy contents of a cld_out_struct to another; */
/* NO memory allocation is done here!!!         */
/************************************************/

int cp_cld_out (cld_out_struct *target, cld_out_struct source) 
{
  int iv=0, lev=0;

  if (target->nlev != source.nlev) {
    fprintf (stderr, "Error: cp_cld_out(), nlev differs between source and target\n");
    fprintf (stderr, "       source: %d,  target: %d\n", source.nlev, target->nlev);
    return -1;
  }

  if (target->nlam != source.nlam) {
    fprintf (stderr, "Error: cp_cld_out(), nlam differs between source and target\n");
    fprintf (stderr, "       source: %d,  target: %d\n", source.nlam, target->nlam);
    return -1;
  }

  target->newsiz = source.newsiz;

  /* copy wavelength- and altitude-dependent stuff */
  for (lev=0; lev<source.nlev-1; lev++) {
    for (iv=0; iv<source.nlam; iv++) {
      target->dtau [iv][lev] = source.dtau [iv][lev];
      target->gg   [iv][lev] = source.gg   [iv][lev];
      target->ssa  [iv][lev] = source.ssa  [iv][lev];
    }
  }
  
  /* copy altitude-dependent stuff */
  for (lev=0; lev<source.nlev; lev++) {
    target->zd   [lev] = source.zd   [lev];
    target->lwc  [lev] = source.lwc  [lev];
    target->effr [lev] = source.effr [lev];
  }
  
  return 0;
}
