/************************************************************************/
/* phase.c                                                              */
/*                                                                      */
/* Calculate phase function from moments                                */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*         DLR, bernhard.mayer@dlr.de                                   */
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "../src_c/ascii.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif


#define PROGRAM "phase"
#define VERSION "0.8a"


/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void print_usage (char *filename)
{
  fprintf (stderr, "%s %s - evaluate phase function\n\n", PROGRAM, VERSION);
  fprintf (stderr, "written by  Bernhard Mayer,\n");
  fprintf (stderr, "            DLR, Oberpfaffenhofen, Germany.\n");
  fprintf (stderr, "            e-Mail bernhard.mayer@dlr.de\n");
  fprintf (stderr, "Version %s finished July 26, 2000\n\n", VERSION);
  fprintf (stderr, "Be aware that this program is an beta version! Please report any\n");
  fprintf (stderr, "kind of error to the author, including the error message and the\n");
  fprintf (stderr, "data being processed when the error occured. Thanks!\n\n");
  fprintf (stderr, "USAGE: %s [options] <filename>\n\n", filename);
  fprintf (stderr, "Optional arguments:\n");
  fprintf (stderr, "  -h            display help message\n");
  fprintf (stderr, "  -c            1-column input\n");
  fprintf (stderr, "  -s <step>     step width for evaluation (default: 0.01)\n\n");
  fprintf (stderr, "%s calculates the phase function p(mu) from the moments of the phase\n", PROGRAM);
  fprintf (stderr, "for an equidistant mu grid.\n");
  fprintf (stderr, "The format of the input file is as generated by the \"mie\" program\n");
  fprintf (stderr, "(the first 7 columns are ignored, the following columns are\n");
  fprintf (stderr, " assumed to hold the moments of the phase function). If option\n");
  fprintf (stderr, "\"-c\" is specified, the input file is considered a one column file,\n");
  fprintf (stderr, "holding one moment per line.\n");
}





/*************************************************************/
/* parse command line options                                */
/*************************************************************/

static int get_options (int argc, char **argv,
			char *programname,
			char *infilename,
			char *onecolumn,
			double *step)
{
  int c=0;
  char *dummy=NULL;

  /* save name of program */
  strncpy (programname, argv[0], FILENAME_MAX);

  /* set defaults */
  strcpy (infilename, "");
  *onecolumn=0;
  *step=0.01;

  while ((c=getopt (argc, argv, "hcs:")) != EOF)  {
    switch(c)  {
    case 'h':  /* help */
      print_usage (programname);
      return (-1);
      break;
    case 'c':  /* one column input */
      *onecolumn=1;
      break;
    case 's':  /* step width*/
      *step = strtod (optarg, &dummy);
      break;
    default:
      print_usage (programname);
      return (-1);
    }
  }
      

  /* check number of remaining command line arguments */
  if (argc - optind != 1)  {
    print_usage (programname);
    return -1;
  }
  
  strncpy (infilename, argv[optind+0], FILENAME_MAX);

  return 0;  /* if o.k. */
}






int main(int argc, char **argv)
{
  int status=0, l=0;

  double *f=NULL;
  
  double sum=0, factor=0;

  double x=0, pl=0, plm1=0, plm2=0;

  int L=0;

  double step=0;
  
  char onecolumn=0;

  char programname[FILENAME_MAX], infilename[FILENAME_MAX]="";

  double **value=NULL;
  int rows=0, min_columns=0, max_columns=0;

  /* get command line options */
  status = get_options (argc, argv, programname, infilename, &onecolumn, &step);

  if (status!=0)
    return status;



  if (onecolumn) {
    status = read_1c_file (infilename, &f, &L);
  
    fprintf (stderr, " ... read %d coefficients from %s\n", L, infilename);
    
    if (status!=0) {
      fprintf (stderr, "error %d reading %s\n", status, infilename);
      return status;
    }
  }
  else {
    status = ASCII_file2double (infilename,    
				&rows,       
				&max_columns,
				&min_columns,
				&value);     

    if (status!=0) {
      fprintf (stderr, "error %d reading %s\n", status, infilename);
      return status;
    }

    if (min_columns!=max_columns)
      fprintf (stderr, "ATTENTION: inconsistent number of columns, %d - %d\n",
	       min_columns, max_columns);
    
    L = min_columns-7;

    fprintf (stderr, " ... read %d lines with %d coefficients each from %s\n", 
	     rows, L, infilename);

    if (rows>1)
      fprintf (stderr, " ... processing only 1st line!\n");


    /* copy first line to work array */
    f = calloc (L, sizeof(double));
    for (l=0; l<L; l++)
      f[l] = value[0][l+7];
  }



  /* multiply by (2l+1) to convert from 'moments of the phase function' */
  /* to Legendre coefficients;                                          */
  for (l=0; l<L; l++)
    f[l] *= (double) (2*l+1);

  /* normalize */
  factor = 1.0 / f[0];

  for (l=0; l<L; l++)
    f[l] *= factor;

  for (x=-1.0; x<=1.00000001; x+=step) {
    
    plm2 = 1;
    plm1 = x;
  
    sum = plm2*f[0] + plm1*f[1];
 
    for (l=2; l<L; l++) {
      pl = ( (double) (2*l-1) * x * plm1 - (double) (l-1) * plm2) / (double) l;
      
      sum += f[l]*pl;

      plm2=plm1;
      plm1=pl;
    }
    
    fprintf (stdout, "%g %g\n", x, sum);
  }


  return 0;
}


